// #define DEBUG_ENABLE_SANITY_CHECKS 1
// #define DEBUG_DUMP_DATA 1
#define SMALLER_MEMORY_MODEL 1
// #define ONLY_COMPUTE_PATH_LENGTHS 1

/*
 *	Atom program to compute whole program path profiles
 *
 *	Purpose:
 *		Generate data to compare/contrast with Wiggins/Redstone
 *		generated traces.
 *
 *	Written by Rick Gorton 7-Apr-1999
 *	gorton@world.std.com
 *
 *	Copyright (C) 1999, Rick Gorton
 *
 * This ATOM tool is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This ATOM tool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this ATOM tool; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * NOTE:
 *	If you find this tool useful, the author is interested in knowing
 *	how you used it, and what kind of modifications you found useful.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "datatypes.h"

/*
 * Data declarations local to this file
 */
long	Total_Procs;
PROC_TABLE_TYPE		*dyn_proc_count;
long	total_proc_count;

long	Total_Blocks;
BLOCK_TABLE_TYPE	*dyn_block_info;
long	total_block_count;

NODE_TABLE_TYPE		*DynNodeCount;

/*
 * Whole program path profile variables
 */
PATH_LIST_TYPE	Active_List;
PATH_LIST_TYPE	Completed_List;
TEMPORAL_ELT	*WPPP_Last_p;	// WPPP == Whole Program Path Profile
TEMPORAL_ELT	*WPPP_Chain;
long		Total_Paths;	// Overall number of path chains executed
long		Total_Unique_Paths;
void convert_active_to_completed();

#ifdef ONLY_COMPUTE_PATH_LENGTHS
long	dyn_path_length[256];
typedef struct PATH_LENGTH_TYPE {
	long	path_length;
	long	counter;
	void	*next;
	} PATH_LENGTH_TYPE;

PATH_LENGTH_TYPE *VeryLongPaths;
#endif

/*
 * Initialization
 */

void DynamicApplicationInit()
{
  long	i, j;
  char	*tmpstring;

  Active_List.path_chain = NULL;
  Active_List.path_length = 0x0;
#ifndef SMALLER_MEMORY_MODEL
  Active_List.first_block_id = 0x0;
#endif
  Active_List.last_block_id = 0x0;
  Active_List.counter = 0x0;
  Active_List.next_path = NULL;

  Completed_List.path_chain = NULL;
  Completed_List.path_length = 0x0;
#ifndef SMALLER_MEMORY_MODEL
  Completed_List.first_block_id = 0x0;
#endif
  Completed_List.last_block_id = 0x0;
  Completed_List.counter = 0x0;
  Completed_List.next_path = NULL;

  WPPP_Last_p = NULL;
  WPPP_Chain = NULL;

  Total_Paths = 0;
  Total_Unique_Paths = 0;
#ifdef ONLY_COMPUTE_PATH_LENGTHS
  VeryLongPaths = NULL;
  for (i = 0; i < 255; i++)
	dyn_path_length[i] = 0;
#endif
}

/*
 * Print out results
 */
void DumpDynamicData()
{
  long	i, j, path_count;
  PATH_LIST_TYPE	*path_walker;
  TEMPORAL_ELT		*WPPP_walker;

#ifdef ONLY_COMPUTE_PATH_LENGTHS
  fprintf (Inst_Logfile, "*** Whole program path length data ***\n");
  fprintf (Inst_Logfile, "\tLength\tCount\n");
  for (i = 0; i < 255; i++)
    {
    if (dyn_path_length[i] > 0)
	fprintf (Inst_Logfile, "\t%ld\t%ld\n", i, dyn_path_length[i]);
    }
  return;
#endif
	/*
	 * Procedure counts
	 */
  fprintf (Inst_Logfile, "*** Procedure Profile data ***\n");
  fprintf (Inst_Logfile, "A Total of %ld procedures were executed\n",
		total_proc_count);
  fprintf (Inst_Logfile, "Procedure number\tProcedure Count\tProcedure Name\n");
  for (i = 0; i < Total_Procs; i++)
    {
    if (dyn_proc_count[i].counter > 0)
	{
	fprintf (Inst_Logfile, "%8ld\t%16ld\t%s\n",
				i, dyn_proc_count[i].counter,
				dyn_proc_count[i].ProcName);
	}
    }

  fprintf (Inst_Logfile, "A Total of %ld blocks were executed\n",
		total_block_count);
	/*
	 * Block Node counts
	 */
  fprintf (Inst_Logfile, "\n*** Simple Block Profile data ***\n");
  fprintf (Inst_Logfile, "Block number\t\tBlock Count\tBlock Addr\tBlock Insns\n");
  for (i = 0; i < Total_Blocks; i++)
    {
    if (DynNodeCount[i].counter > 0)
	{
	fprintf (Inst_Logfile, "%8ld\t%16ld\t0x%lx\t%ld\n",
				i, DynNodeCount[i].counter,
				dyn_block_info[i].Block_Addr,
				dyn_block_info[i].Num_Insns);
	}
    }

	/*
	 * Whole Program Paths
	 */
	// Clean up the dangling program exit path

  convert_active_to_completed ();
  fprintf (Inst_Logfile, "\n*** Whole Program Paths ***\n");
  fprintf (Inst_Logfile, "A total of %ld Whole Program Paths were executed\n",
		Total_Paths);
  fprintf (Inst_Logfile, "A total of %ld Unique Whole Program Paths were found\n",
		Total_Unique_Paths);

  fprintf (Inst_Logfile, "Unique path data:\n-------------");
  path_walker = &Completed_List;
  for (i = 1; i <= Total_Unique_Paths; i++)
    {
    WHOLE_PATH_BLOCK_ELT	*chain_p;

    fprintf (Inst_Logfile, "\nPath: %ld\n\t", i);
    chain_p = path_walker->path_chain;
    for (j = 0; j < path_walker->path_length; j+= 2)
	{
	fprintf (Inst_Logfile, "\t%ld", chain_p->u.block_id_int[0]);
	if (chain_p->u.block_id_int[1] != 0)
	    fprintf (Inst_Logfile, "\t%ld", chain_p->u.block_id_int[1]);
	if ((j + 2) % 8 == 0)
	    fprintf (Inst_Logfile, "\n\t");
	chain_p = chain_p->next_block;
	}
#ifdef DEBUG_ENABLE_SANITY_CHECKS
    if (chain_p != NULL)
	fprintf (Inst_Logfile, "*** ERROR *** chain_p not null after %ld blocks\n", j);
#endif
    path_walker = path_walker->next_path;
    }

#ifdef DEBUG_ENABLE_SANITY_CHECKS
  if (path_walker != NULL)
    {
    fprintf (Inst_Logfile, "*** ERROR *** path_walker not null after %ld paths\n", i);
    }
#endif

	// And now the order in which the various paths were executed
  fprintf (Inst_Logfile, "\n-------------\nTemporal Super-Path id data:\n");
  WPPP_walker = WPPP_Chain;
  path_count = 0;
  while (WPPP_walker != NULL)
    {
    if (WPPP_walker->sequential_count > 1)
	fprintf (Inst_Logfile, "\t\t%ld (%ld)", WPPP_walker->path_seq_num,
			WPPP_walker->sequential_count);
    else
	fprintf (Inst_Logfile, "\t\t%ld", WPPP_walker->path_seq_num);
    if (path_count % 4 == 3)
	    fprintf (Inst_Logfile, "\n");
    path_count++;
    WPPP_walker = WPPP_walker->next;
    }
  fprintf(Inst_Logfile, "\n*** Dynamic data end ***\n");
}


/*
 * Procedure data collection
 */

void DynAllocateStorage (long Proc_Elts, long Block_Elts)
{
  Total_Procs = Proc_Elts;
  total_proc_count = 0;
  dyn_proc_count = (PROC_TABLE_TYPE *) calloc (sizeof(PROC_TABLE_TYPE), Proc_Elts);

  Total_Blocks = Block_Elts;
  total_block_count = 0;
  dyn_block_info = (BLOCK_TABLE_TYPE *) calloc (sizeof(BLOCK_TABLE_TYPE), Block_Elts);

  DynNodeCount = (NODE_TABLE_TYPE *) calloc (sizeof (NODE_TABLE_TYPE), Block_Elts);
}

void Dyn_ProcCountInit (long PSeq, char *Pname, long Addr)
{
  strncat (dyn_proc_count[PSeq].ProcName, Pname, PROC_NAME_SIZE);
  dyn_proc_count[PSeq].StartAddr = Addr;
}

void Procedure_Count (long ProcIndex)
{
  long	i;

  dyn_proc_count[ProcIndex].counter++;
  total_proc_count++;
}

void BbInfo (long this_BlockNum, long this_ProcNum,
		long this_BlockSize, long this_BlockAddr)
{
	/*
	 * Called once before the program runs, used solely for
	 * cross-correlation of stuff at data output time.
	 */
  dyn_block_info[this_BlockNum].ProcId = this_ProcNum;
  dyn_block_info[this_BlockNum].Num_Insns = this_BlockSize;
  dyn_block_info[this_BlockNum].Block_Addr = this_BlockAddr;
}

	/*
	 * Various path collection code routines
	 */

void DeleteActive_Path ()
{
  WHOLE_PATH_BLOCK_ELT	*discard_elt_p;

#ifdef DEBUG_DUMP_DATA
    fprintf (Inst_Logfile, "DeleteActive_Path\n");
#endif

	// reclaim active path chain elements
  discard_elt_p = Active_List.path_chain;
  while (discard_elt_p != NULL)
    {
    WHOLE_PATH_BLOCK_ELT	*tmp2_p;

    tmp2_p = discard_elt_p->next_block;
    free (discard_elt_p);
    discard_elt_p = tmp2_p;
    }
  Active_List.path_chain = NULL;
}

void Initial_Completion()
{
  TEMPORAL_ELT	*newt_p;

	// Only once - very first completed path.
  Total_Unique_Paths++;

  Completed_List.path_chain = Active_List.path_chain;
  Completed_List.path_length = Active_List.path_length;
#ifndef SMALLER_MEMORY_MODEL
  Completed_List.first_block_id = Active_List.first_block_id;
#endif
  Completed_List.last_block_id = Active_List.last_block_id;
  Completed_List.counter = 1;
  Completed_List.seq_num = 1;
#ifdef DEBUG_ENABLE_SANITY_CHECKS
  if (Total_Unique_Paths != 1)
    {
    fprintf (Inst_Logfile, "Sanity error Total_Unique_Paths initially %ld\n",
		Total_Unique_Paths);
    }
#endif

	// Now we need to append Path_Seq_Id to the temporal chain

  newt_p = malloc (sizeof (TEMPORAL_ELT));
  newt_p->sequential_count = 1;
  newt_p->path_p = &Completed_List;
  newt_p->next = NULL;
  newt_p->path_seq_num = 1;	// It will be 0x1, anyhow
  WPPP_Chain = newt_p;
  WPPP_Last_p = newt_p;
  return;
}

PATH_LIST_TYPE *gen_path_sequence()
{
  PATH_LIST_TYPE	*completed_walker, *back_one;
  long			got_a_match;
  long			Path_Seq_Id;

    got_a_match = 0;
    completed_walker = &Completed_List;
    back_one = completed_walker;
    Path_Seq_Id = 0;
    while ((completed_walker != NULL) && !got_a_match)
	{
	Path_Seq_Id++;
	if ((completed_walker->path_length == Active_List.path_length)
#ifndef SMALLER_MEMORY_MODEL
		&& (completed_walker->first_block_id == Active_List.first_block_id)
#endif
		&& (completed_walker->last_block_id == Active_List.last_block_id))
	    {
		// length, first, last blocks match. check entire path_chain
		// for exact match
	    WHOLE_PATH_BLOCK_ELT	*new_chain_p, *curr_chain_p;
	    long	i;

	    curr_chain_p = completed_walker->path_chain;
	    new_chain_p = Active_List.path_chain;
	    for (i = 0; i < Active_List.path_length; i+=2)
		{
		// Compare two nodes at a time (much faster)
		if (curr_chain_p->u.block_id_long != new_chain_p->u.block_id_long)
		    {
		    break;
		    }
		curr_chain_p = curr_chain_p->next_block;
		new_chain_p = new_chain_p->next_block;
		}

	    if (i >= Active_List.path_length)
		{
		// Exact Match
#ifdef DEBUG_ENABLE_SANITY_CHECKS
		if ((curr_chain_p != NULL) || (new_chain_p != NULL))
		    fprintf (Inst_Logfile, "Sanity error: convert... chain walking\n");
#endif
		got_a_match = 1;
		completed_walker->counter++;
		DeleteActive_Path ();
		return completed_walker;
		}
	    else	// Didn't match - try the next active one
		{
		    back_one = completed_walker;
		    completed_walker = completed_walker->next_path;
		}	// length, first, last block ids match
	    }
	else
	   {	// First, Last, and length not the same
	   back_one = completed_walker;
	   completed_walker = completed_walker->next_path;
	   }
	}

    if (!got_a_match)
	{
	PATH_LIST_TYPE	*newp_p;

#ifdef DEBUG_ENABLE_SANITY_CHECKS
	if (completed_walker != NULL)
	    fprintf (Inst_Logfile, "non-null completed_walker = 0x%lx\n");
#endif
	Total_Unique_Paths++;

	newp_p = malloc (sizeof (PATH_LIST_TYPE));
	newp_p->path_chain = Active_List.path_chain;
	newp_p->path_length = Active_List.path_length;
#ifndef SMALLER_MEMORY_MODEL
	newp_p->first_block_id = Active_List.first_block_id;
#endif
	newp_p->last_block_id = Active_List.last_block_id;
	newp_p->counter = 1;
	newp_p->seq_num = Total_Unique_Paths;
	newp_p->next_path = NULL;
#ifdef DEBUG_ENABLE_SANITY_CHECKS
	if (Total_Unique_Paths != Path_Seq_Id + 1)
	    fprintf (Inst_Logfile, "gen_path... Total %ld != Path_Seq %ld\n",
				Total_Unique_Paths, Path_Seq_Id);
	if (back_one->next_path != NULL)
	    fprintf (Inst_Logfile, "gen_path... back_one->next == 0x%lx\n",
				back_one->next_path);
#endif
	back_one->next_path = newp_p;
	return newp_p;
	}
#ifdef DEBUG_ENABLE_SANITY_CHECKS
    fprintf (Inst_Logfile, "ASSERT error: match, but didn't return earlier\n");
#endif
}


#ifdef ONLY_COMPUTE_PATH_LENGTHS
	/*
	 * All we want to do is figure out how long the majority of the
	 * paths are.  This data will enable the development of more efficient
	 * (in terms of memory) compression techniques.
	 */
void convert_active_to_completed ()
{

  if (Active_List.path_length > 0xff)
    {
    fprintf (Inst_Logfile, "Very Long Path seen: %ld\n", Active_List.path_length);
    }
  else
    {
    dyn_path_length[Active_List.path_length]++;
    }
  DeleteActive_Path();
	// Clean up Active_List
  Active_List.path_chain = NULL;
  Active_List.path_length = 0x0;
#ifndef SMALLER_MEMORY_MODEL
  Active_List.first_block_id = 0x0;
#endif
  Active_List.last_block_id = 0x0;
  Active_List.counter = 0x0;
  Active_List.next_path = NULL;
}

#else
	/* Not ONLY_COMPUTE_PATH_LENGTHS */
void convert_active_to_completed ()
{
  PATH_LIST_TYPE	*new_path_p;
  long			got_a_match;
  long			Path_Seq_Id;

#ifdef DEBUG_DUMP_DATA
    fprintf (Inst_Logfile, "convert_active_to_completed\n");
    fflush (Inst_Logfile);
#endif
    Total_Paths++;
    got_a_match = 0;	// False
	// look for a match against prior paths, and get a path id
    if (Completed_List.path_chain == NULL)
	{
	Initial_Completion();
	return;
	}

	/*
	 * Not the initial entry.  Get the sequence number of its path
	 */
    new_path_p = gen_path_sequence();

	// Extend the existing temporal chain if necessary
    if (WPPP_Last_p->path_seq_num == new_path_p->seq_num)
	{
	WPPP_Last_p->sequential_count++;
	}
    else
	{
	TEMPORAL_ELT	*newt_p;

	newt_p = malloc (sizeof (TEMPORAL_ELT));
	newt_p->path_seq_num = new_path_p->seq_num;
	newt_p->sequential_count = 1;
	newt_p->path_p = new_path_p;
	newt_p->next = NULL;

	WPPP_Last_p->next = newt_p;
	WPPP_Last_p = newt_p;

#ifdef DEBUG_ENABLE_SANITY_CHECKS
#ifdef DEBUG_DUMP_DATA
    {
    TEMPORAL_ELT	*wppp_p;

    wppp_p = WPPP_Chain;
    fprintf (Inst_Logfile, "WPPP_Chain is:\n");
    while (wppp_p != NULL)
	{
	PATH_LIST_TYPE	*tmp_path_p;

	tmp_path_p = wppp_p->path_p;
	fprintf (Inst_Logfile, "\t\t%ld (%ld) last block: %ld\n",
			wppp_p->path_seq_num,
			wppp_p->sequential_count,
			tmp_path_p->last_block_id);
	fflush (Inst_Logfile);
	wppp_p = wppp_p->next;
	}
    }
#endif
#endif
	}

	// Clean up Active_List
    Active_List.path_chain = NULL;
    Active_List.path_length = 0x0;
#ifndef SMALLER_MEMORY_MODEL
    Active_List.first_block_id = 0x0;
#endif
    Active_List.last_block_id = 0x0;
    Active_List.counter = 0x0;
    Active_List.next_path = NULL;

}
	/* endif is for ONLY_COMPUTE_PATH_LENGTHS */
#endif

void Extend_Path (long this_BlockNum)
{
  WHOLE_PATH_BLOCK_ELT		*elt_p, *chain_walker, *back_one;
  long			found;
  long			i;

#ifdef DEBUG_DUMP_DATA
    {
    fprintf (Inst_Logfile, "Extend_Path (%ld) @ 0x%lx\n",
		this_BlockNum, dyn_block_info[this_BlockNum].Block_Addr);
    fflush (Inst_Logfile);
    }
#endif
    total_block_count++;
    DynNodeCount[this_BlockNum].counter++;

    if (Active_List.path_chain == NULL)
	{
	elt_p = malloc (sizeof (WHOLE_PATH_BLOCK_ELT));
	elt_p->u.block_id_int[0] = this_BlockNum;
	elt_p->u.block_id_int[1] = 0;
	elt_p->next_block = NULL;

	// Very first block in this sequence; simple extension
	Active_List.path_chain = elt_p;
	Active_List.path_length = 1;
#ifndef SMALLER_MEMORY_MODEL
	Active_List.first_block_id = this_BlockNum;
#endif
	Active_List.last_block_id = this_BlockNum;
	Active_List.next_path = NULL;
	return;
	}

	/*
	 * Walk the existing chain, and check for a loop (block id match)
	 * If we found a match, we have completed a path.
	 */

    found = 0;
    chain_walker = Active_List.path_chain;

    back_one = chain_walker;
    for (i = 0; i < Active_List.path_length; i+= 2)
	{
	if ((chain_walker->u.block_id_int[0] == this_BlockNum)
		|| (chain_walker->u.block_id_int[1] == this_BlockNum))
	    {
	    found = 1;
	    break;	// Quit the loop
	    }
	back_one = chain_walker;
	chain_walker = chain_walker->next_block;
	}

#ifdef DEBUG_ENABLE_SANITY_CHECKS
    if ((chain_walker != NULL) && !found)
	{
	fprintf (Inst_Logfile, "Extend: mismatch on length walk: i == %ld\n", i);
	}
#endif

    if (found)
	{
	convert_active_to_completed ();

	// Now start a new path profile with this node as the first entry
	elt_p = malloc (sizeof (WHOLE_PATH_BLOCK_ELT));
	elt_p->u.block_id_int[0] = this_BlockNum;
	elt_p->u.block_id_int[1] = 0;
	elt_p->next_block = NULL;

	Active_List.path_chain = elt_p;
	Active_List.path_length = 1;
#ifndef SMALLER_MEMORY_MODEL
	Active_List.first_block_id = this_BlockNum;
#endif
	Active_List.last_block_id = this_BlockNum;
	Active_List.next_path = NULL;
	}
    else
	{
	// Simply extend this path, as it hasn't been seen yet on the
	// current active path profile.
	// Since we are compressing two nodes into a long, we only
	// need to allocate a new chain node if the current one is full.
	if (back_one->u.block_id_int[1] == 0)
	    back_one->u.block_id_int[1] = this_BlockNum;
	else
	    {
	    elt_p = malloc (sizeof (WHOLE_PATH_BLOCK_ELT));
	    elt_p->u.block_id_int[0] = this_BlockNum;
	    elt_p->u.block_id_int[1] = 0;
	    elt_p->next_block = NULL;
	    back_one->next_block = elt_p;
	    }
	Active_List.path_length++;

	Active_List.last_block_id = this_BlockNum;
	}
}
